package ch.hevs.kart;

/**
 * Interface to implement if an object is interested in basic kart status changes. An object 
 * implementing this interface can be added using the Kart's method
 * {@link Kart#addKartListener(KartListener)}. The object will get notified about changes of the
 * most important parameters of the Kart.
 * <br><br>
 * Note that not all changes in the registers of the kart are reported by this listener interface.
 * If an object needs more details about all the registers of the Kart, it should implement the 
 * {@link KartStatusRegisterListener} interface and subscribe himself using the method
 * {@link Kart#addStatusRegisterListener(KartStatusRegisterListener)}.
 * 
 * @author Nils Chatton
 * @author Christopher Metrailler (mei@hevs.ch)
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
@SuppressWarnings("UnusedParameters")
public interface KartListener {
	/**
	 * Called by the kart whenever the actual steering position has changed.
	 * <br><br>
	 * This method is part of the KartListener interface and all listeners get notified when the 
	 * value of the actual steering position has changed.
	 * 
	 * @param kart					The kart object.
	 * @param absolutePosition		The actual position from 0 to max steps (depends kart model).
	 * @param normalizedPosition	The actual position from -1 (left) to 1 (right).
	 */
	void steeringPositionChanged( Kart kart, int absolutePosition, float normalizedPosition);
	
	/**
	 * The steering system has changed its state.
	 * <br><br>
	 * This method is called when either the steering position has just reached its end position 
	 * (reached = true) or when the steering system has got a new target value and starts the 
	 * move to this new position (reached = false).
	 * 
	 * @param kart		The kart object.
	 * @param reached	True if the steering target position has reached, false otherwise.
	 */
	void steeringPositionReachedChanged(Kart kart, boolean reached);
	
	/**
	 * The battery level has changed.
	 * <br><br>
	 * This method is called by the kart object whenever the battery level on the kart changes.
	 * 
	 * @param kart		The kart object.
	 * @param level		The battery level from 0 (empty) to 1 (full).
	 */
	void batteryVoltageChanged(Kart kart, float level);

    /**
     * A sequence that was started using {@link Kart#run(KartSequence)} has been completed.
     * <br><br>
     * <b>IMPORTANT: This feature is experimental and might not be implemented on all Karts!</b>
     *
     * @param kart      The kart object.
     */
    void sequenceCompleted(Kart kart);

	/**
	 * The connection status to the Kart has changed.
	 * <br><br>
	 * This method is called if the Bluetooth connection to the kart has been closed or is broken.
	 * 
	 * @param kart		The kart object.
	 * @param online	True if the kart is online, false otherwise.
	 */
	void connectionStatusChanged(Kart kart, boolean online);
	
	/**
	 * The kart has a message to display, most of the time this are cryptic messages just for
	 * developers, but they might be helpful to display during development...
	 * 
	 * @param kart		The kart object.
	 * @param message	The message.
	 */
	void message(Kart kart, String message);
}
