package ch.hevs.kart;

import android.content.Context;

import org.xmlpull.v1.XmlPullParser;
import org.xmlpull.v1.XmlPullParserException;
import org.xmlpull.v1.XmlPullParserFactory;

import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.util.LinkedList;
import java.util.List;

import ch.hevs.utils.Limits;

/**
 * A KartSequence is a step-by-step sequence of commands that can be downloaded to the Kart.
 * <br><br>
 * You can see the hardware based sequence player on the Kart as a very simple virtual machine that
 * executes one instruction after another and supports a well defined set of opcodes along with
 * their arguments. This objects abstracts the hard work to generate these vm commands by exposing a
 * simple interface to construct the sequence and using the method {@link Kart#run(KartSequence)}
 * you can execute such a sequence on the Kart.
 * <br><br>
 * The builder pattern ({@link KartSequence.Builder}) is used in order to construct such a sequence.
 * Sequences can be saved and loaded to/from xml files by using the methods
 * {@link KartSequence#save} and {@link KartSequence#load}.
 * <br><br>
 * <b>IMPORTANT: This feature is experimental and might not be implemented on all Karts!</b>
 *
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.2 - 2016
 */
public class KartSequence {
    /**
     * Creates and returns a new Builder in order to construct a Kart sequence. See the
     * {@link KartSequence.Builder} documentation for available commands a sequence can be made of.
     *
     * @return  New KartSequence Builder.
     */
    public Builder Builder() {
        return new Builder();
    }

    /**
     * Builder that must be used in order to construct new KartSequences. Every method returns a
     * reference to the builder himself, so it is possible to create sequences in just one line of
     * code.
     *
     * <code>
     *     KartSequence seq = KartSequence.Builder().steer(0).drive(3).during(5000);
     * </code>
     *
     * The example above lets the Kart drive in a circle for about 3 seconds.
     */
    public static class Builder {
        /**
         * This instruction does actually nothing. It is present for completeness.
         *
         * @return  Reference to the builder object.
         */
        public Builder nop() {
            sequenceToConstruct.instructions.add(new Instruction(Instruction.OpCode.Nothing, (short)0));
            return this;
        }

        /**
         * This instruction changes the speed of the kart. Valid values for the speed are -15..15,
         * whereas 0 means stop, -15 is the fastest speed backwards and 15 is the fastest speed in
         * the forward direction.
         *
         * @param speed Speed to set (-15..15).
         * @return      Reference to the builder object.
         */
        public Builder drive(int speed) {
            speed = Limits.limit(speed, Kart.MINIMAL_SPEED, Kart.MAXIMAL_SPEED);
            sequenceToConstruct.instructions.add(
                    new Instruction(Instruction.OpCode.SetDriveSpeed, (short)speed));
            return this;
        }

        /**
         * This instruction changes the steering to the given position. Valid values for the
         * command are 0..4095, but the highest value depends the actual geometry of the Kart.
         * There is no control at all if the given parameter is compatible with the steering
         * mechanism, so it is your duty to ensure that the values are acceptable.
         *
         * @param position  Position to set steering (0..4095, but depends steering geometry).
         * @return          Reference to the builder object.
         */
        public Builder steer(final int position) {
            sequenceToConstruct.instructions.add(
                    new Instruction(Instruction.OpCode.SetSteeringPosition, (short)position));
            return this;
        }

        /**
         * This instruction sets the four LED outputs to the values of the four least significant
         * bytes of the given value. All other bits are ignored.
         *
         * @param values    Value to set the four outputs (LED) to.
         * @return          Reference to the builder object.
         */
        public Builder outputs(int values) {
            values = values & 0xF;
            sequenceToConstruct.instructions.add(
                    new Instruction(Instruction.OpCode.SetDigitalOutputs, (short)values));
            return this;
        }

        /**
         * This instruction will turn the LED at the given index (0..3) on or generally spoken, the
         * digital output will be activated.
         *
         * @param index Index of the LED/digital output to turn on (0..4).
         * @return      Reference to the builder object.
         */
        public Builder ledOn(final int index) {
            if (index >= 0 && index < 4) {
                sequenceToConstruct.instructions.add(
                        new Instruction(Instruction.OpCode.SetLed, (short)index));
                return this;
            } else {
                throw new IndexOutOfBoundsException(
                        "There are only four LEDs/outputs on the kart (0-3)");
            }
        }

        /**
         * This instruction will turn the LED at the given index (0..3) off or generally spoken, the
         * digital output will be deactivated.
         *
         * @param index Index of the LED/digital output to turn off (0..4).
         * @return      Reference to the builder object.
         */
        public Builder ledOff(final int index) {
            if (index >= 0 && index < 4) {
                sequenceToConstruct.instructions.add(
                        new Instruction(Instruction.OpCode.ClearLed, (short)index));
                return this;
            } else {
                throw new IndexOutOfBoundsException(
                        "There are only four LEDs/outputs on the kart (0-3)");
            }
        }

        /**
         * This instruction will leave the Kart in the exact same state until the given amount of
         * pulses are counted on hall sensor 1. If the hall sensor counter register corresponds to
         * the number of rotations the axis has made, you can make the kart traveling a certain
         * distance with the current steering position and speed. The number of instructions can be between
         * 0 and 4095.
         * <br><br>
         * Setting the speed {@link KartSequence.Builder#drive(int)} to 0 before this instruction
         * might not be a smart move...
         *
         * @param pulseCount    Number of hall sensor 1 pulses to wait until the next action
         *                      (0..4095).
         * @return              Reference to the builder object.
         */
        public Builder travel(final int pulseCount) {
            sequenceToConstruct.instructions.add(
                    new Instruction(Instruction.OpCode.WaitHall1Pulses, (short)pulseCount));
            return this;
        }

        /**
         * This instruction will leave the Kart in the exact same state for the given amount of
         * milliseconds. This method of waiting is certainly less exact in terms of the absolute
         * position than the {@link KartSequence.Builder#travel(int)} instruction, but might be
         * still useful for simple tasks.
         * <br><br>
         * Note that the actual resolution of the times is 10ms. A delay of 0ms up to about 41s
         * can be achieved by providing values from 0 (0ms) to 40950 (40.95s). If the passed value
         * is bigger than 40950, the delay will be limited to 41s.
         *
         * @param milliseconds  Number of milliseconds to wait before executing the next
         *                      instruction.
         * @return              Reference to the builder object.
         */
        public Builder during(int milliseconds) {
            milliseconds = Limits.limit(milliseconds, 0, 40950);
            sequenceToConstruct.instructions.add(
                    new Instruction(Instruction.OpCode.WaitMilliseconds, (short)(milliseconds / 10)));
            return this;
        }

        /**
         * Creates and returns the actual kart sequence. Note that the builder can be reused in
         * order to create a new sequence.
         *
         * @return  Built kart sequence.
         */
        public KartSequence build() {
            sequenceToConstruct.instructions.add(new Instruction(Instruction.OpCode.End, (short)0));

            KartSequence sequence = sequenceToConstruct;
            sequenceToConstruct = new KartSequence();
            return sequence;
        }

        // Only KartSequence class can instantiate a builder object.
        private Builder() {}

        // The actual sequence in construction.
        private KartSequence sequenceToConstruct = new KartSequence();
    }

    // Instruction class, represents a single instruction (opcode + operand).
    private static class Instruction {
        // The opcodes actually supported by the Kart'a sequence vm.
        public enum OpCode {
            // NOP. Probably not used that often...
            Nothing((byte)0),

            // Set the speed of the driving subsystem.
            SetDriveSpeed((byte)1),

            // Set the steering position.
            SetSteeringPosition((byte)2),

            // Set digital output port to the given value (4 lsb).
            SetDigitalOutputs((byte)3),

            // Turn the LED at the given index on.
            SetLed((byte)10),

            // Turn the LED at the given index off.
            ClearLed((byte)11),

            // Wait for the given amount of hall pulses on hall sensor 1.
            WaitHall1Pulses((byte)4),

            // Wait for the given amount of milliseconds before executing the next instruction.
            WaitMilliseconds((byte)5),

            // End of the sequence (Added by the builder automatically) and starts the sequence.
            End((byte)0xF);

            OpCode(final byte opCode) {
                this.opCode = opCode;
            }

            public byte toByte() {
                return opCode;
            }

            private byte opCode;
        }

        public Instruction(final OpCode opCode, final short operand) {
            this.opCode = opCode;
            this.operand = operand;
        }

        public int getKartVMInstruction() {
            int instruction = opCode.toByte() << 12;
            instruction |= operand & 0x0FFF;
            return instruction;
        }

        public String toXml() {
            return String.format("<Instruction opcode=\"%s\" operand=\"%s\"/>",
                    opCode.toString(), operand);
        }

        private OpCode opCode;
        private short operand;
    }

    public int[] getKartVMInstructions() {
        int[] instructions = new int[this.instructions.size()];
        for (int i = 0; i < instructions.length; ++i) {
            instructions[i] = this.instructions.get(i).getKartVMInstruction();
        }
        return instructions;
    }

    private List<Instruction> instructions = new LinkedList<Instruction>();

    public void save(final String filename, final Context context) throws IOException {
        FileOutputStream output = context.openFileOutput(filename, Context.MODE_PRIVATE);
        output.write("<KartSequence>\n".getBytes());
        for (Instruction instruction: instructions) {
            output.write(("    " + instruction.toXml() + "\n").getBytes());
        }
        output.write("</KartSequence>\n".getBytes());
        output.close();
    }

    public static KartSequence load(final String filename, final Context context)
            throws IOException, XmlPullParserException {
        KartSequence.Builder builder = new KartSequence.Builder();
        FileInputStream input = context.openFileInput(filename);
        XmlPullParser parser = XmlPullParserFactory.newInstance().newPullParser();
        parser.setInput(input, null);
        int eventType = parser.getEventType();
        while (eventType != XmlPullParser.END_DOCUMENT) {
            switch (eventType) {
                case XmlPullParser.START_TAG:
                    if ("Instruction".equals(parser.getName())) {
                        Instruction.OpCode opcode = null;
                        Short operand = null;
                        for (int i = 0; i < parser.getAttributeCount(); ++i) {
                            if ("opcode".equals(parser.getAttributeName(i))) {
                                try {
                                    opcode = Instruction.OpCode.valueOf(parser.getAttributeValue(i));
                                } catch (IllegalArgumentException e) {
                                    e.printStackTrace();
                                    throw new IOException("Invalid opcode: " + opcode);
                                }
                            } else if ("operand".equals(parser.getAttributeName(i))) {
                                operand = Short.parseShort(parser.getAttributeValue(i));
                            }
                        }
                        if (opcode != null && operand != null) {
                            switch (opcode) {
                                case Nothing:
                                    builder.nop();
                                    break;

                                case SetDriveSpeed:
                                    builder.drive(operand);
                                    break;

                                case SetSteeringPosition:
                                    builder.steer(operand);
                                    break;

                                case SetDigitalOutputs:
                                    builder.outputs(operand);
                                    break;

                                case SetLed:
                                    builder.ledOn(operand);
                                    break;

                                case ClearLed:
                                    builder.ledOff(operand);
                                    break;

                                case WaitHall1Pulses:
                                    builder.travel(operand);
                                    break;

                                case WaitMilliseconds:
                                    builder.during(operand);
                                    break;
                            }
                        }
                    }
                    break;
            }
            eventType = parser.next();
        }
        return builder.build();
    }
}
