package ch.hevs.kart.comm.protocol.kartv1;

import java.nio.ByteBuffer;

import ch.hevs.kart.KartControlRegister;
import ch.hevs.kart.KartStatusRegister;
import ch.hevs.kart.comm.protocol.Protocol;
import ch.hevs.kart.comm.protocol.ProtocolListener;
import ch.hevs.kart.comm.protocol.ProtocolState;
import ch.hevs.kart.comm.protocol.UiThreadProtocolListenerWrapper;
import ch.hevs.kart.comm.transport.Transport;
import ch.hevs.kart.comm.transport.TransportState;

/**
 * Version 1 of the protocol between the kart and the Java device.
 *
 * @author Michael Clausen (clm@hevs.ch)
 * @version 2.1 - 2015
 */
public final class KartV1Protocol implements Protocol {
	@Override
	public void initialize(final Transport transport, final ProtocolListener listener) {
		this.transport = transport;
		this.listener = new UiThreadProtocolListenerWrapper();
		this.listener.setListener(listener);
	}

	@Override
	public void dispose() {
		if (transport == null) return;
		
		transport.disconnect();
		transport = null;
		listener.setListener(null);
	}
	
	@Override
	public boolean isReady() {
		return transport != null && transport.isConnected();
	}
	
	@Override
	public void setRegisterValue(final KartControlRegister register, final int value) {
		if (transport == null) return;
		
		ByteBuffer frame = ByteBuffer.allocate(5);
		frame.put((byte)0x55);
		frame.put((byte)register.getAddress());
		frame.putShort((short)value);
		frame.put((byte)0xAA);
		
		transport.write(frame);
	}

	
	/*** Communication with kart via transport layer **********************************************/
	
	// Actual dissection state. 
	private enum DissectionState {
		SoF, Address, Data, EoF
	}

	@Override
	public void dataIndication(final Transport transport, final ByteBuffer buffer) {
		// Reception FSM...
		while(buffer.hasRemaining()) {
			switch (readState) {
				case SoF:
					if (buffer.get() == 0x55) readState = DissectionState.Address;
					break;
					
				case Address:
					actualDissectionAddress = buffer.get();
					readState = DissectionState.Data;
					break;
					
				case Data:
					if (buffer.remaining() >= 2) {
						actualDissectionValue = buffer.getShort();
						readState = DissectionState.EoF;
					} else 
						readState = DissectionState.SoF;
					break;
					
				case EoF:
					byte b = buffer.get();
					if (b == (byte)0xAA) {
						if (listener != null) {
							KartStatusRegister register = 
									KartStatusRegister.fromAddress(actualDissectionAddress);
							if (register != null)
								listener.registerValueIndication(this, register, 
															     actualDissectionValue);
						} 
							
					}
					readState = DissectionState.SoF;
					break;
			}
		}
	}

	@Override
	public void stateChangeIndication(final Transport transport, final TransportState state) {
		if (transport == this.transport)
			setState(state == TransportState.Connected ? 
					 ProtocolState.Online : ProtocolState.Offline);
	}

	@Override
	public void errorIndication(final Transport transport, final String error) {
		if (transport == this.transport && listener != null)
			listener.errorIndication(this, "Transport error: " + error);
	}

	private Transport transport;
	private DissectionState readState = DissectionState.SoF;
	private byte actualDissectionAddress;
	private short actualDissectionValue;
	

	/*** Listener management **********************************************************************/

	private void setState(final ProtocolState state) {
		if(this.state != state) {
			this.state = state;
			listener.stateChangeIndication(this, state);
		}
	}

	private UiThreadProtocolListenerWrapper listener;
	
	private ProtocolState state = ProtocolState.Offline;
}
